package Font::Encoding;
require 5.00404;
use strict;

# Font::Encoding -- manipulate font encodings for TeX
# Copyright (C) 1997-2001  Melissa O'Neill
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 02111-1307, USA.


BEGIN {
    use Exporter   ();
    use vars       qw($VERSION @ISA @EXPORT @EXPORT_OK);
    $VERSION     = 0.01;
    @ISA         = qw(Exporter);
    @EXPORT      = qw();
    @EXPORT_OK   = qw();
}

use Carp;
use integer;

# use vars @EXPORT_OK;

#
# $glyphencoding->{ Name    => <string:encodingname>,
#		    ByCode  => { <int:charcode> => <string:glyph>,... },
#		    ByGlyph => { <string:glyph> => [ <int:charcode>,... ],...}}
#
# (the ByGlyph portion is optional, created on demand)
#

sub new ($) {
    @_ == 1 or croak 'usage: $enc = Font::Encoding->new();';
    my ($thing) = @_;
    my $class = ref($thing) || $thing;
    my $glyphencoding = { Name   => 'UnspecifiedEncoding',
    			  ByCode => {} };
    return bless $glyphencoding, $class;
}

sub clone ($;$) {
    @_ == 1 or @_ == 2 or croak 'usage: $enc2 = $enc1->clone( [ NEWNAME ] );';
    my ($orig, $name) = @_;
    $name = $orig->{Name} unless defined $name;
    my $class = ref($orig);
    my $clone = { Name   => $name,
    	          ByCode => { %{$orig->{ByCode}} } };
    $clone->{ByGlyph} = { %{$orig->{ByGlyph}} } if defined $orig->{ByGlyph};
    return bless $clone, $class;
}

sub newFromFile ($$) {
    @_ == 2 or croak 'usage: $enc = Font::Encoding->newFromFile( FILENAME );';
    my ($thing, $encfilename) = @_;
    my ($count, $invector) = (0, 0);
    my %encoding = ();
    my $encodingname;
    open (ENCODING, "$encfilename") or croak "Can't open '$encfilename'";
    while (<ENCODING>) {
	chomp;
	s/%.*//s;  # remove comments
	s/^\s+//s; # remove leading whitespace
	s/\s+$//s; # remove trailing whitespace
	next if m/^$/;
#	print "Line = '$_'\n";
	s/([\[\]])/ $1 /;
	foreach my $symbol (split) {
#	    print "Processing '$symbol'\n";
	      if ($symbol eq '[') {
		  croak "Encoding not named before '['" unless defined $encodingname;
		  $invector++;
	      } elsif ($symbol eq ']') {
		  $invector--;
	      } elsif ($invector) {
		  croak "Not a valid PostScript name" unless $symbol =~ s|^/||;
		  if ($symbol ne '.notdef') {
		    $encoding{$count} = $symbol;
		  }
		  $count++;
	      } elsif (!$invector and $symbol ne 'def') {
		  croak "Unexpected word '$symbol' " if defined $encodingname;
		  croak "Not a valid PostScript name" unless $symbol =~ s|^/||;
		  $encodingname = $symbol;
	      }
	}
    }
    close ENCODING;
    croak "Encoding vector has $count elements, should be 256" if $count != 256;

    my $glyphencoding = { Name   => $encodingname,
    			  ByCode => \%encoding };

    my $class = ref($thing) || $thing;
    return bless $glyphencoding, $class;
}

sub _invertHash ($) {
    my ($inhash) = @_;
    my %outhash = ();
    while (my ($key, $value) = each %{$inhash}) {
	if (exists $outhash{$value}) {
	    push(@{$outhash{$value}}, $key);
	} else {
	    $outhash{$value} = [$key]
	}
    }
    return \%outhash;
}

sub name ($) {
    @_ == 1 or croak 'usage: $enc->name();';
    my ($glyphencoding) = @_;
    return $glyphencoding->{Name};
}

sub renameTo ($$) {
    @_ == 2 or croak 'usage: $enc->renameTo( ENCODINGNAME );';
    my ($glyphencoding, $encname) = @_;
    $glyphencoding->{Name} = $encname;
    return $glyphencoding;
}

sub glyphFor ($$) {
    @_ == 2 or croak 'usage: $enc->glyphFor( CHARCODE );';
    my ($glyphencoding, $charcode) = @_;
    return $glyphencoding->{ByCode}->{$charcode};
}

sub _forceByGlyph ($) {
    my ($glyphencoding) = @_;
    my $byglyph = $glyphencoding->{ByGlyph};
    return $byglyph if defined $byglyph;
    return $glyphencoding->{ByGlyph} = _invertHash($glyphencoding->{ByCode});
}

sub charcodesFor ($$) {
    @_ == 2 or croak 'usage: @charcodes = $enc->charcodesFor( GLYPHNAME );';
    my ($glyphencoding, $glyph) = @_;
    my $charcodes;
    
    my $byglyph = _forceByGlyph $glyphencoding;
    $charcodes = $byglyph->{$glyph};
    $charcodes = \@$charcodes;
    return @$charcodes;
}

sub encodeAs ($$$) {
    @_ == 3 or croak 'usage: $enc->encodeas( CHARCODE, GLYPHNAME );';
    my ($glyphencoding, $charcode, $glyph) = @_;

    # Squark loudly if we try to encode on character position as two
    # different glyphs. We also return if we're trying to encode a
    # position that's already encoded the same.
    
    if (defined (my $priorglyph = $glyphencoding->{ByCode}->{$charcode})) {
        if ($priorglyph eq $glyph) {
	    return;
	} else {
	    carp "Redefining $glyphencoding->{Name}\[$charcode\] from '$priorglyph' to '$glyph'";
	    delete $glyphencoding->{ByGlyph}; # Easiest way...
	}
    }

    # Add the character to the ByCode part.
    
    $glyphencoding->{ByCode}->{$charcode} = $glyph;

    # Add the character to the ByGlyph part, if it exists. We could just
    # delete the ByGlyph part and let it regenerate naturally. 
    if (defined (my $byglyph = $glyphencoding->{ByGlyph})) {
	if (defined $byglyph->{$glyph}) {
	    push(@{$byglyph->{$glyph}}, $charcode);
	} else {
	    $byglyph->{$glyph} = [$charcode]
	}
    }
}

sub deleteCharcodes ($@) {
    @_ > 1 or croak 'usage: $enc->deleteCharcodes( CHARCODE, ... );';
    my ($glyphencoding, @deletia) = @_;

    foreach my $charcode (@deletia) {
	delete $glyphencoding->{ByCode}->{$charcode};
    }
    delete $glyphencoding->{ByGlyph};
}

sub glyphs ($) {
    @_ == 1 or croak 'usage: @glyphs = $enc->glyphs();';
    my ($glyphencoding) = @_;

    _forceByGlyph($glyphencoding);
    return keys %{$glyphencoding->{ByGlyph}};
}

sub charcodes ($) {
    @_ == 1 or croak 'usage: @glyphs = $enc->charcodes();';
    my ($glyphencoding) = @_;

    return keys %{$glyphencoding->{ByCode}};
}

sub renameGlyph($$$) {
    @_ == 3 or croak 'usage: @glyphs = $enc->renameGlyph( OLDNAME, NEWNAME );';
    
    my ($glyphencoding, $oldglyphname, $newglyphname) = @_;

    _forceByGlyph($glyphencoding);
    my $charcodes = delete $glyphencoding->{ByGlyph}->{$oldglyphname};
    return $glyphencoding unless defined $charcodes;
    $glyphencoding->{ByGlyph}->{$newglyphname} = $charcodes;
    foreach my $charcode (@$charcodes) {
        $glyphencoding->{ByCode}->{$charcode} = $newglyphname;
    }
}


sub _renameHash {
    my $hash = shift;
    my %newname = @_;
    foreach my $key (keys %$hash) {
	delete $hash->{$key} unless $newname{$key};
    }
    while (@_) {
        my $data  = delete $hash->{(shift)};
	my $key   = shift;
	$hash->{$key} = $data if defined $data;
    }
}

sub restrictAndRenameGlyphs ($%) {
    (@_ >= 3 and (@_ % 2) == 1)
	  or croak 'usage: $enc->restrictGlyphs( GLYPH, NEWNAME ... );';
    
    my $glyphencoding = shift;
    my $byglyph = _forceByGlyph($glyphencoding);
    _renameHash $byglyph, @_;
    my %bycode;
    while (my ($glyph, $charcodes) = each %$byglyph) {
	foreach my $charcode (@$charcodes) {
	    $bycode{$charcode} = $glyph;
	}
    }
    $glyphencoding->{ByCode} = \%bycode;
    return $glyphencoding;
}



sub empties ($) {
    @_ == 1 or croak 'usage: @glyphs = $enc->charcodes();';
    my ($glyphencoding) = @_;
    my @empties;
    my $bycode = $glyphencoding->{ByCode};
    
    foreach my $charcode (0..255) {
        push @empties, $charcode unless defined $bycode->{$charcode}
    }
    return @empties;
}

sub writeEnc ($\%) {
    @_ == 1 or @_ == 2 or croak 'usage: $enc->writeEnc( [ FILENAME ] );';
    my ($glyphencoding, $encfile) = @_;
    my $bycode = $glyphencoding->{ByCode};
    my $encname = $glyphencoding->name;
    $encfile = "$encname.enc" unless defined $encfile;
    open ENCFILE, ">$encfile" or croak "Can't open '$encfile'";
    print ENCFILE "/$encname [";
    for my $charcode (0..255) {
        if ($charcode % 8) {
	  print ENCFILE " ";
	} else {
	  print ENCFILE "\n";
	}
	my $glyph = $bycode->{$charcode};
	$glyph = ".notdef" unless defined $glyph;
	print ENCFILE "/$glyph";
    }
    print ENCFILE "\n] def\n";
    close ENCFILE or croak "Couldn't close '$encfile'";
}

sub writeEncLongwinded ($\%) {
    @_ == 1 or @_ == 2 or croak 'usage: $enc->writeEnc( [ FILENAME ] );';
    my ($glyphencoding, $encfile) = @_;
    my $bycode = $glyphencoding->{ByCode};
    my $encname = $glyphencoding->name;
    $encfile = "$encname.enc" unless defined $encfile;
    open ENCFILE, ">$encfile" or croak "Can't open '$encfile'";
    print ENCFILE "/$encname [\n";
    for my $charcode (0..255) {
	my $glyph = $bycode->{$charcode};
	$glyph = ".notdef" unless defined $glyph;
	$glyph = "/$glyph";
	format ENCFILE =
  @<<<<<<<<<<<<<<<<<<<<<% @<<<<<
  $glyph				, sprintf('%02X', $charcode)
.
	write ENCFILE;
    }
    print ENCFILE "] def\n";
    close ENCFILE or croak "Couldn't close '$encfile'";
}
