package Font::MMdesc;

# Font::MMdesc -- grok the many names and properties that MM fonts may have
# Copyright (C) 1997-2001  Melissa O'Neill
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 02111-1307, USA.


use strict;
use Carp;

BEGIN {
    use Exporter   ();
    use vars       qw($VERSION @ISA @EXPORT @EXPORT_OK);
    $VERSION     = 0.01;
    @ISA         = qw(Exporter);
    @EXPORT      = qw();
    @EXPORT_OK   = qw();
}

sub _invertHash ($;$) {
    my ($inhash, $outhash) = @_;
    while (my ($key, $value) = each %{$inhash}) {
	push(@{$outhash->{$value}}, $key);
    }
    return $outhash;
}


#
# Different worlds use different terms to mean essentially the same thing.
# Thus, a ``Light'' weight for a font may be ``LT'' if we're using the
# Adobe nomenclature for fonts, 'l' in the LaTeX scheme, 'l' in the
# Karl Berry scheme, etc.
#
# Internally, we use long, hopefully meaningful, names.

# WEIGHT TABLES

# Karl Berry scheme

my %toKB_Weight = (
    UltraLight	=> 'a',
    ExtraLight	=> 'j',
    Light	=> 'l',
    Book	=> 'k',
    Regular	=> 'r',
    Semibold	=> 's',
    Demi	=> 'd',
    Bold	=> 'b',
    ExtraBold	=> 'x',
    Black	=> 'c'
);

my %fromKB_Weight = reverse %toKB_Weight;

# NFSS scheme

my %toNFSS_Weight = (
    UltraLight	=> 'ul',
    ExtraLight	=> 'el',
    Light	=> 'l',
    Book	=> 'k',
    Regular	=> 'm',
    Semibold	=> 'sb',
    Demi	=> 'd',
    Bold	=> 'b',
    ExtraBold	=> 'eb',
    Black	=> 'ub'
);
my %fromNFSS_Weight = reverse %toNFSS_Weight;

# Multiple Master scheme

my %toMM_Weight = (
    Light	=> 'LT',
    Book	=> 'BK',
    Regular	=> 'RG',
    Semibold	=> 'SB',
    Demi	=> 'DM',
    Bold	=> 'BD',
    Black	=> 'BL'
);

my %fromMM_Weight = reverse %toMM_Weight;

# WIDTH TABLES

# Karl Berry scheme

my %toKB_Width = (
    UltraCondensed	=> 'o',
    ExtraCondensed	=> 'q',
    Condensed		=> 'c',
    SemiCondensed	=> 'n',
    Normal		=> '',
    SemiExtended	=> 's',
    Extended		=> 'x',
    ExtraExtended	=> 'v',
    UltraExtended	=> 'w'
);

my %fromKB_Width = reverse %toKB_Width;

# NFSS scheme

my %toNFSS_Width = (
    UltraCondensed	=> 'uc',
    ExtraCondensed	=> 'ec',
    Condensed		=> 'c',
    SemiCondensed	=> 'sc',
    Normal		=> 'r',
    SemiExtended	=> 'sx',
    Extended		=> 'x',
    ExtraExtended	=> 'ex',
    UltraExtended	=> 'ux'
);
my %fromNFSS_Width = reverse %toNFSS_Width;

# Multiple Master scheme

my %toMM_Width = (
    Condensed	    => 'CN',
    SemiCondensed   => 'SC',
    Normal	    => 'NO',
    SemiExtended    => 'SE',
    Extended	    => 'EX'
);

my %fromMM_Width = reverse %toMM_Width;

# SHAPE TABLES

# Karl Berry scheme

my %toKB_Shape = (
    Italic		=> 'i',
    Alternate		=> 'a',
    Normal		=> 'r',
    Oblique		=> 'o',
    Ornaments		=> 'p',
    Outline		=> 'l',
    OldStyleFigures	=> 'j',
    TitlingFigures	=> 'b',
    Shadow		=> 'h',
    SmallCaps		=> 'c',
    FakeSmallCaps	=> 'g',		# Really code for SmallText
    Swash		=> 'w',
    UprightItalic	=> 'u',
);
my %fromKB_Shape = reverse %toKB_Shape;

my @seqKB_Shape = qw(TitlingFigures SmallCaps FakeSmallCaps Shadow Italic Oblique OldStyleFigures UprightItalic Swash Ornaments Alternate);

# NFSS scheme

my %toNFSS_Shape = (
    Italic		=> 'it',
    Normal		=> 'n',
    Oblique		=> 'sl',
#   Outline		=> '??',
#   Shadow		=> '??',
    SmallCaps		=> 'sc',
    Swash		=> 'sw',
    UprightItalic	=> 'up',
);
my %fromNFSS_Shape = reverse %toNFSS_Shape;
my @seqNFSS_Shape = qw(Swash Italic UprightItalic Oblique SmallCaps);


# Multiple Master scheme

my %toMM_Shape = (
    Italic		=> 'It',
    Normal		=> '',
    Alternate		=> 'Alt',
    Oblique		=> 'Oblique',
    Ornaments		=> 'Or',
    SmallCaps		=> 'SC',
    Swash		=> 'Sw',
    Expert		=> 'Ep',
);
my %fromMM_Shape = reverse %toMM_Shape;
my @seqMM_Shape = qw(Swash Italic Oblique Alternate SmallCaps Ornaments Expert);


# FONTNAME TABLES

# Karl Berry scheme

my %fromKB_FontName = (
    'paj' => 'AJensonMM',
    'pkc' => 'CronosMM',
    'pkp' => 'KeplMM',
    'pmn' => 'MinionMM',
    'pmy' => 'MyriadMM',
    'pne' => 'NuevaMM',
    'ptk' => 'TektonMM'
);

my %toKB_FontName = reverse %fromKB_FontName;


# ENCODING/VARIANT TABLES

my %fromKB_Encoding = (
    '8r' => 'TeXBase1Encoding',
    '8x' => 'ExpertEncoding',
    '8a' => 'AdobeStandardEncoding',
    '0y' => 'TeXnANSIEncoding',
    '1y' => 'TeXnANSIEncoding',
    '8y' => 'TeXnANSIEncoding',
    '9y' => 'TeXnANSIEncoding'
);

my %fromKB_Variant = (
    '8r' => 'Regular',
    '8x' => 'Regular',
    '8a' => 'Regular',
    '0y' => 'Inferior',
    '1y' => 'Superior',
    '8y' => 'Regular',
    '9y' => 'Expertized'
);

my %toKB_EncVar;
foreach my $kb (keys %fromKB_Encoding) {
    $toKB_EncVar{$fromKB_Encoding{$kb}}{$fromKB_Variant{$kb}} = $kb;
}

# INFORMATION ABOUT MULTIPLE MASTER FONTS

my %macMM_genericAxisNames = (
    Weight	=> 'wt',
    Width	=> 'wd',
    OpticalSize => 'op'
);

my %macMM_specificAxisNames = (
    Weight	=> \%toMM_Weight,
    Width	=> \%toMM_Width,
    OpticalSize => 'OP'
);


my %amfm = (
    KeplMM => {
        FamilyName => "Kepler MM",
	MasterFonts => { map { $_ => 1 } qw(
	    KeplMM KeplMM-It KeplMM-Ep KeplMM-ItEp KeplMM-ItSC KeplMM-SC
	    KeplMM-Sw
	)},
	InitFunc => sub {
	    my ($fontdesc) = @_;
	    my $ornaments = $fontdesc->{Ornaments};
	    if ($ornaments) {
		$fontdesc->{BaseName} = "KeplMM-Or";
	    }
	},
	Axes => [qw(Weight Width OpticalSize)],
	Weight => {
	    Light	=> 275,
	    Book	=> 325,
	    Regular	=> 385,
	    Semibold	=> 535,
	    Demi	=> 590,
	    Bold	=> 650,
	    Black	=> 775
	},
	Width => {
	    Condensed	    => 450,
	    SemiCondensed   => 510,
	    Normal	    => 575,
	    SemiExtended    => 710,
	    Extended	    => 850
	},
	Primary => [
	    [OpticalSize => 10, Weight => 'Light', Width => 'Normal'],
	    [OpticalSize => 10, Weight => 'Regular', Width => 'Condensed'],
	    [OpticalSize => 7, Weight => 'Regular', Width => 'Normal'],
	    [OpticalSize => 10, Weight => 'Regular', Width => 'Normal'],
	    [OpticalSize => 72, Weight => 'Regular', Width => 'Normal'],
	    [OpticalSize => 10, Weight => 'Regular', Width => 'Extended'],
	    [OpticalSize => 10, Weight => 'Semibold', Width => 'Normal'],
	    [OpticalSize => 10, Weight => 'Bold', Width => 'Normal'],
	    [OpticalSize => 10, Weight => 'Black', Width => 'Normal']
	]
    },
   'KeplMM-Or' => {
        FamilyName => "Kepler MM",
	MasterFonts => { map { $_ => 1 } qw(
	    KeplMM-Or1 KeplMM-Or2 KeplMM-Or3
	)},
	Ornaments => sub {
	   my ($value, $shape, $fontdesc) = @_;
	   return "Or$value";
	},
	Axes => [qw(OpticalSize)],
	Primary => [
	    [OpticalSize => 10],
	    [OpticalSize => 72]
	]
    },
    MinionMM => {
        FamilyName => "Minion MM",
	MasterFonts => { map { $_ => 1 } qw(
	    MinionMM MinionMM-It MinionMM-Ep MinionMM-ItEp MinionMM-ItSC
	    MinionMM-SC MinionMM-Sw
	)},
	InitFunc => sub {
	    my ($fontdesc) = @_;
	    my $ornaments = $fontdesc->{Ornaments};
	    if ($ornaments) {
		$fontdesc->{BaseName} = "MinionMM-Or";
	    }
	},
	Axes => [qw(Weight Width OpticalSize)],
	Weight => {
	    Book	=> 345,
	    Regular	=> 367,
	    Semibold	=> 485,
	    Demi	=> 540,
	    Bold	=> 578,
	    ExtraBold   => 620
	},
	Width => {
	    Condensed	    => 465,
	    SemiCondensed   => 525,
	    Normal	    => 585,
	    SemiExtended    => 600
	},
	Primary => [
	    [ OpticalSize => 11, Weight => 'Regular', Width => 'Condensed'],
	    [ OpticalSize => 11, Weight => 'Regular', Width => 'Normal'],
	    [ OpticalSize => 72, Weight => 'Regular', Width => 'Normal'],
	    [ OpticalSize => 11, Weight => 'Semibold', Width => 'Condensed'],
	    [ OpticalSize => 11, Weight => 'Semibold', Width => 'Normal'],
	    [ OpticalSize => 11, Weight => 'Bold', Width => 'Condensed'],
	    [ OpticalSize => 11, Weight => 'Bold', Width => 'Normal']
	]
    },
   'MinionMM-Or' => {
        FamilyName => "Minion MM",
	MasterFonts => { map { $_ => 1 } qw(
	    MinionMM-Or
	)},
	Axes => [qw(OpticalSize)],
	Primary => [
	    [ OpticalSize => 11],
	    [ OpticalSize => 72]
	]
    },
    AJensonMM => {
        FamilyName => "Adobe Jenson MM",
	MasterFonts => { map { $_ => 1 } qw(
	    AJensonMM AJensonMM-Alt AJensonMM-Ep
	    AJensonMM-It AJensonMM-ItAlt AJensonMM-ItEp 
	    AJensonMM-ItSC AJensonMM-SC AJensonMM-Sw
	)},
	Axes => [qw(Weight OpticalSize)],
	Weight => {
	    Book	=> 340, # Not figured out yet
	    Regular	=> 371,
	    Semibold	=> 506,
	    Demi	=> 540,	# Not figured out yet
	    Bold	=> 590,
	    ExtraBold   => 620
	},
	Primary => [
	    [ OpticalSize => 8,  Weight => 'Regular'],
	    [ OpticalSize => 12, Weight => 'Regular'],
	    [ OpticalSize => 72, Weight => 'Regular'],
	    [ OpticalSize => 8,  Weight => 'Semibold'],
	    [ OpticalSize => 12, Weight => 'Semibold'],
	    [ OpticalSize => 72, Weight => 'Semibold'],
	    [ OpticalSize => 8,  Weight => 'Bold'],
	    [ OpticalSize => 12, Weight => 'Bold'],
	    [ OpticalSize => 72, Weight => 'Bold']
	]
    },
    MyriadMM => {
        FamilyName => "Myriad MM",
	MasterFonts => { map { $_ => 1 } qw(
	    MyriadMM-It MyriadMM
	)},
	Axes => [qw(Weight Width)],
	Weight => {
	    Light	=> 215,
	    Book	=> 340,
	    Regular	=> 400,
	    Semibold	=> 565,
	    Demi	=> 610,
	    Bold	=> 700,
	    Black	=> 830
	},
	Width => {
	    Condensed	    => 300,
	    SemiCondensed   => 450,
	    Normal	    => 600,
	    SemiExtended    => 700,
	},
	FakeSmallCaps => {
	    Weight => {
		Light	    => 285,
		Book	    => 450,
		Regular	    => 521.25,
		Semibold    => 730,
	    	Demi	    => 792
	    },
	    Width => {
		Normal	    => 675
	    },
	    Scale => 59/80,
	    Tracking => 35 # 25
	},
	Primary => [
	    [ Weight => 'Light', Width => 'Condensed'],
	    [ Weight => 'Light', Width => 'Normal'],
	    [ Weight => 'Light', Width => 'SemiExtended'],
	    [ Weight => 'Regular', Width => 'Condensed'],
	    [ Weight => 'Regular', Width => 'Normal'],
	    [ Weight => 'Regular', Width => 'SemiExtended'],
	    [ Weight => 'Semibold', Width => 'Condensed'],
	    [ Weight => 'Semibold', Width => 'Normal'],
	    [ Weight => 'Semibold', Width => 'SemiExtended'],
	    [ Weight => 'Bold', Width => 'Condensed'],
	    [ Weight => 'Bold', Width => 'Normal'],
	    [ Weight => 'Bold', Width => 'SemiExtended'],
	    [ Weight => 'Black', Width => 'Condensed'],
	    [ Weight => 'Black', Width => 'Normal'],
	    [ Weight => 'Black', Width => 'SemiExtended']
	]
    },
    CronosMM => {
        FamilyName => "Cronos MM",
	MasterFonts => { map { $_ => 1 } qw(
	    CronosMM-Ep CronosMM-It CronosMM-ItEp CronosMM-ItSC CronosMM-Or
	    CronosMM-SC CronosMM-Sw CronosMM
	)},
	Axes => [qw(Weight OpticalSize)],
	Weight => {
	    Light	=> 250,
	    Book	=> 345,
	    Regular	=> 408,
	    Semibold	=> 574,
	    Demi	=> 620,
	    Bold	=> 700
	},
	Primary => [
	    [ OpticalSize => 11, Weight => 'Light'],
	    [ OpticalSize => 72, Weight => 'Light'],
	    [ OpticalSize => 11, Weight => 'Regular'],
	    [ OpticalSize => 72, Weight => 'Regular'],
	    [ OpticalSize => 11, Weight => 'Semibold'],
	    [ OpticalSize => 72, Weight => 'Semibold'],
	    [ OpticalSize => 11, Weight => 'Bold'],
	    [ OpticalSize => 72, Weight => 'Bold']
	]
    },
    NuevaMM => {
        FamilyName => "Nueva MM",
	MasterFonts => { map { $_ => 1 } qw(
	    NuevaMM-It NuevaMM
	)},
	Axes => [qw(Weight Width)],
	Weight => {
	    Light	=> 200,
	    Book	=> 260,
	    Regular	=> 330,
	    Demi	=> 425,
	    Bold	=> 540,
	    ExtraBold	=> 580
	},
	Width => {
	    Condensed	    => 200,
	    SemiCondensed   => 337.5,
	    Normal	    => 475,
	    SemiExtended    => 612.5,
	    Extended	    => 750,
	},
	FakeSmallCaps => {
	    Weight => {
		Light	    => 260,
		Book	    => 340,
		Regular	    => 420,
		Demi	    => 540
	    },
	    Width => {
		SemiCondensed	=> 400,
		Normal		=> 560,
		SemiExtended	=> 725
	    },
	    Scale => 59/80,
	    Tracking => 30
	},
	Primary => [
	    [Weight => 'Light', Width => 'Condensed'],
	    [Weight => 'Light', Width => 'Normal'],
	    [Weight => 'Light', Width => 'Extended'],
	    [Weight => 'Regular', Width => 'Condensed'],
	    [Weight => 'Regular', Width => 'Normal'],
	    [Weight => 'Regular', Width => 'Extended'],
	    [Weight => 'Bold', Width => 'Condensed'],
	    [Weight => 'Bold', Width => 'Normal'],
	    [Weight => 'Bold', Width => 'Extended']
	]
    },
    TektonMM => {
        FamilyName => "Tekton MM",
	MasterFonts => { map { $_ => 1 } qw(
	    TektonMM-Oblique TektonMM
	)},
	Axes => [qw(Weight Width)],
	Weight => {
	    Light	=> 100,
	    Regular	=> 240,
	    Semibold	=> 325,
	    Bold	=> 503,
	    Black	=> 620
	},
	Width => {
	    Condensed	    => 250,
	    SemiCondensed   => 410,
	    Normal	    => 564,
	    SemiExtended    => 710,
	    Extended	    => 850
	},
	FakeSmallCaps => {
	    Weight => {
		Light	    => 145,
		Regular	    => 312.5,
	        Semibold    => 425,
	        Bold	    => 620	# Should actually be about 640, but
					# 620 is the max and looks okay.
	    },
	    Width => {
		SemiCondensed	=> 480,
		Normal		=> 658,
		SemiExtended	=> 790
	    },
	    Scale => 59/80,
	    Tracking => 15
	},
	Primary => [
	    [Weight => 'Light', Width => 'Condensed'],
	    [Weight => 'Light', Width => 'Normal'],
	    [Weight => 'Light', Width => 'Extended'],
	    [Weight => 'Regular', Width => 'Condensed'],
	    [Weight => 'Regular', Width => 'Normal'],
	    [Weight => 'Regular', Width => 'Extended'],
	    [Weight => 'Bold', Width => 'Condensed'],
	    [Weight => 'Bold', Width => 'Normal'],
	    [Weight => 'Bold', Width => 'Extended']
	]
    }
);    


sub master_fontname ($) {
    my ($fontdesc) = @_;
    my $basename = $fontdesc->{BaseName};
    my $master_font = $basename;
    $master_font =~ s/MM-.*/MM/;
    my $fontdata = $amfm{$basename};
    my @variants;
    foreach my $shape (@seqMM_Shape) {
        my $value = $fontdesc->{$shape};
        if ($value) {
	    my $special = $fontdata->{$shape};
	    my $word;
	    if (defined $special) {
		if (ref $special) {
		    $word = &$special($value, $shape, $fontdesc);
		} else {
		    $word = $special;
		}
	    } else {
		$word = $toMM_Shape{$shape};
	    }
	    push (@variants, $word);
	}
    }
    $master_font .= "-" . join('', @variants) if (@variants);
    return $master_font;
}

sub instance_fontname ($) {
    my ($fontdesc) = @_;
    my $instance_font = master_fontname $fontdesc;
    my @blend;
    my $basename = $fontdesc->{BaseName};
    my $fontdata = $amfm{$basename};
    my $axes = $fontdata->{Axes};
    if ($fontdesc->{FakeSmallCaps}) {
        $fontdata = $fontdata->{FakeSmallCaps};
	croak "No FakeSmallCaps entry for $basename" unless defined $fontdata;
    } 
    foreach my $axis (@$axes) {
        my $value = $fontdesc->{$axis};
	my $param = $fontdata->{$axis};
	$param = defined $param ? $param->{$value} : $value;
	croak "No value for $axis/$value" unless defined $param;
	push @blend, $param;
    }
    $instance_font .= '_' . join('_', @blend) . '_';
    return $instance_font;
}

sub mac_fontname ($) {
    my ($fontdesc) = @_;
    my $mac_font = master_fontname $fontdesc;
    my @blend;
    my $fontdata = $amfm{$fontdesc->{BaseName}};
    my $isprimary = $fontdesc->is_primary;
    foreach my $axis (@{$fontdata->{Axes}}) {
        my $value = $fontdesc->{$axis};
	my $param = $fontdata->{$axis};
	$param = defined $param ? $param->{$value} : $value;
	my $code = $macMM_genericAxisNames{$axis};
	if ($isprimary) {
	    $code = $macMM_specificAxisNames{$axis};
	    $code = $code->{$value} if ref $code;
	} else {
	    $code = $macMM_genericAxisNames{$axis};
	}
	push @blend, $param, $code;
    }
    $mac_font .= '_' . join(' ', @blend);
    return $mac_font;
}

sub alias_fontname ($) {
    my ($fontdesc) = @_;
    my $basename = $fontdesc->{BaseName};
    my $head = $basename;
    $head =~ s/MM-.*/MM/;
    my $opsize = $fontdesc->{OpticalSize};
    $head .= $opsize . 'pt' if defined $opsize;
    my $tail = '';
    my $fontdata = $amfm{$basename};
    foreach my $axis (@{$fontdata->{Axes}}) {
	next unless defined $fontdata->{$axis};
	my $word = $fontdesc->{$axis};
	croak "No value for axis" unless defined $word;
	next if $word eq 'Normal';
#	$word = substr($word, 0, 3) if length $word > 3;
	$tail .= $word;
    }
    foreach my $shape (@seqMM_Shape) {
        my $value = $fontdesc->{$shape};
        if ($value) {
	    my $special = $fontdata->{$shape};
	    my $word;
	    if (defined $special) {
		if (ref $special) {
		    $word = &$special($value, $shape, $fontdesc);
		} else {
		    $word = $special;
		}
		$word =~ s/\Q$toMM_Shape{$shape}\E/$shape/;
	    } else {
		$word = $shape;
	    }
	    $tail .= $word;
	}
    }
    $tail =~ s/SmallCaps/SC&OSF/;
#    $tail =~ s/([A-Z][a-z][a-z])[a-z]+/$1/g;
    return "$head-$tail";
}

sub alias_family ($) {
    my ($fontdesc) = @_;
    my $alias_family =  $amfm{$fontdesc->{BaseName}}{FamilyName};
    my $opsize = $fontdesc->{OpticalSize};
    $alias_family .= " ${opsize}pt" if defined $opsize;
    return $alias_family;
}

sub alias_fullname ($) {
    my ($fontdesc) = @_;
    my $alias_fullname = alias_family $fontdesc;
    my $fontdata = $amfm{$fontdesc->{BaseName}};
    foreach my $axis (@{$fontdata->{Axes}}) {
	next unless defined $fontdata->{$axis};
	my $word = $fontdesc->{$axis};
	croak "No value for axis" unless defined $word;
	next if $word eq 'Normal';
	$alias_fullname .= " $word";
    }
    foreach my $shape (@seqMM_Shape) {
        my $value = $fontdesc->{$shape};
        if ($value) {
	    my $special = $fontdata->{$shape};
	    my $word;
	    if (defined $special) {
		if (ref $special) {
		    $word = &$special($value, $shape, $fontdesc);
		} else {
		    $word = $special;
		}
		$word =~ s/\Q$toMM_Shape{$shape}\E/$shape/;
	    } else {
		$word = $shape;
	    }
	    $alias_fullname .= " $word";
	}
    }
    return $alias_fullname;
}

sub kb_fontname_with_enc ($;$) {
    my ($fontdesc, $encoding) = @_;
    my $basefont = $fontdesc->{BaseName};
    $basefont =~ s/MM-.*/MM/;
    my $kb_fontname = $toKB_FontName{$basefont};
    $kb_fontname .= $toKB_Weight{$fontdesc->{Weight}};
    croak "No Weight char for $fontdesc->{Weight}" unless defined $toKB_Weight{$fontdesc->{Weight}};
    foreach my $shape (@seqKB_Shape) {
        my $value = $fontdesc->{$shape};
        $kb_fontname .= $toKB_Shape{$shape} x $value if $value; 
    }
    $encoding = ($fontdesc->{Expert} ? 'ExpertEncoding'
		  : 'AdobeStandardEncoding') unless defined $encoding;
    my $variant  = $fontdesc->{Variant};
    $variant = 'Regular' unless defined $variant;
    my $kb_encvar = $toKB_EncVar{$encoding}{$variant};
    croak "No KB fontname for $encoding/$variant" unless defined $kb_encvar;
    $kb_fontname .= $kb_encvar;
    my $width = $fontdesc->{Width};
    $kb_fontname .= $toKB_Width{$width} if defined $width;
    my $opsize = $fontdesc->{OpticalSize};
    return $kb_fontname unless defined $opsize;
    return sprintf("%s%04d", $kb_fontname, int($opsize * 100 + 0.5));
}

sub kb_fontname ($) {
    my ($fontdesc) = @_;
    return kb_fontname_with_enc($fontdesc, $fontdesc->{Encoding});
}


sub kb_instance_pfa($) {
    my ($fontdesc) = @_;
    return kb_fontname_with_enc($fontdesc);
}

sub kb_pfa ($) {
    my ($fontdesc) = @_;
    my $basefont = $fontdesc->{BaseName};
    my ($hiding_fakesc, $fakesc);
    if (exists $fontdesc->{FakeSmallCaps}) {
	# Hide our fake small-caps-ness
        ++$hiding_fakesc;
	$fakesc = delete $fontdesc->{FakeSmallCaps};
    }
    $basefont =~ s/MM-.*/MM/;
    my $kb_fontname = $toKB_FontName{$basefont};
    $kb_fontname .= 'z';
    foreach my $shape (@seqKB_Shape) {
        my $value = $fontdesc->{$shape};
        $kb_fontname .= $toKB_Shape{$shape} x $value if $value; 
    }
    $kb_fontname .= $fontdesc->{Expert} ? '8x' : '8a';
    $fontdesc->{FakeSmallCaps} = $fakesc if ($hiding_fakesc);
    return $kb_fontname;
}

sub weight ($) {
    my ($fontdesc) = @_;
    return $fontdesc->{Weight};
}

sub encoding ($) {
    my ($fontdesc) = @_;
    return $fontdesc->{Encoding};
}

sub small_fakedata ($) {
    my ($fontdesc) = @_;
    croak "Not a Small font" unless $fontdesc->{FakeSmallCaps};
    my $basename = $fontdesc->{BaseName};
    my $fontdata = $amfm{$basename};
    $fontdata = $fontdata->{FakeSmallCaps};
    croak "No FakeSmallCaps entry for $basename" unless defined $fontdata;
    return ($fontdata->{Tracking}, $fontdata->{Scale});
}

sub primaries ($) {
    my ($fontdesc) = @_;
    my $basename;
    if (ref $fontdesc) {
	$basename = $fontdesc->{BaseName}
    } else {
        $basename = $fontdesc;
    }
    my $fontdata = $amfm{$basename};
    return unless defined $fontdata;
    return $fontdata->{Primary};
}

sub is_primary {
    my ($fontdesc) = @_;
    my $primaries = $fontdesc->primaries;
    PRIMARY: foreach my $primary (@$primaries) {
        my @bits = @$primary;
	while (@bits) {
	    my ($key, $value) = (shift @bits, shift @bits);
#	    print "$key: $value vs. $fontdesc->{$key}\n";
	    next PRIMARY if ($fontdesc->{$key} ne $value);
	}
	return 1;
    }
    return 0;
}

	# This doesn't do much.
sub is_valid {
    my ($fontdesc) = @_;
    my $psfontname = $fontdesc->master_fontname;
    my $fontdata = $amfm{$fontdesc->{BaseName}};
#   print "Trying to find $psfontname in [",
#	join(', ', keys %{$fontdata->{MasterFonts}}),"]\n";
    return 0 unless defined $fontdata;
    return '' if ($fontdesc->{Small} and !$fontdata->{FakeSmallCaps});
    return $fontdata->{MasterFonts}{$psfontname};
}

sub new {
    my ($thing, $object) = @_;
    my $class = ref($thing) || $thing;
    bless \%{$object}, $class;
}

sub copy {
    my ($fontdesc) = @_;
    my %copy = %{$fontdesc};
    bless \%copy, ref $fontdesc;
}

sub newFromKB {
    my ($thing, $kb_fontname) = @_;
    my $class = ref($thing) || $thing;
    my ($kfontname, $kweight, $kshapes, $kencvar, $kwidth, $kopsize) =
        $kb_fontname =~ m/^(...)(.)(\D*)(..)(\D*)(\d*)/
    or croak "Can't parse $kb_fontname as a Karl Berry fontname";
    my $basename = $fromKB_FontName{$kfontname};
    croak "Don't know anything about $kfontname" unless defined $basename;
    my $fontdesc = {
        BaseName => $basename,
	Weight   => $fromKB_Weight{$kweight},
	Width    => $fromKB_Width{$kwidth}
    };
    if ($kencvar eq '8x') {
        ++$fontdesc->{Expert};
    } elsif ($kencvar ne '8a') {
	$fontdesc->{Encoding} = $fromKB_Encoding{$kencvar};
	$fontdesc->{Variant}  = $fromKB_Variant{$kencvar};
    }
    foreach my $key (keys %$fontdesc) {
        croak "Can't parse $kb_fontname as a Karl Berry fontname (invalid $key)"
	    unless defined $fontdesc->{$key};
    }
    foreach my $kshape (split //, $kshapes) {
        my $shape = $fromKB_Shape{$kshape};
	croak "Can't parse $kb_fontname as a Karl Berry: can't grok '$kshape' as a shape" unless defined $shape;
	++$fontdesc->{$shape};
    }
    if ($kopsize ne '') {
        $kopsize /= 100;
	$fontdesc->{OpticalSize} = $kopsize;
    }
    bless $fontdesc, $class;
    my $fontdata = $amfm{$basename};
    if (defined $fontdata and defined (my $initfunc = $fontdata->{InitFunc})) {
        &$initfunc($fontdesc);
	$basename = $fontdesc->{BaseName};
	$fontdata = $amfm{$basename};
    }
    croak "No AMFM information about $basename" unless defined $fontdata;
    my %hasAxis = map { $_ => 1 } @{$fontdata->{Axes}};
    my %checkedAxes = (
        Weight		=> 'Regular',
	Width		=> 'Normal',
	OpticalSize	=> undef
    );
    while (my ($axis, $default) = each %checkedAxes) {
	my $value = $fontdesc->{$axis};
        if ($hasAxis{$axis}) {
	    croak "KB fontname `$kb_fontname' specifies no value for $axis axis" 
		unless defined $value;
	    my $axisdata = $fontdata->{$axis};
	    if (defined $axisdata) {
		croak "$basename has no $axis value for $value"
		    unless defined $axisdata->{$value};
	    }
	} else {
	    if (defined $default) {
	        croak "$basename has no $axis axis -- $axis must be $default (not $value)"
		     unless $value eq $default;
	    } else {
	        croak "$basename has no $axis axis, but $axis = $value was specified"
		     if defined $value;
	    }
	}
    }
    foreach my $shape (@seqMM_Shape) {
        my $value   = $fontdesc->{$shape};
        my $special = $fontdata->{$shape};
	croak "$shape has invalid value $value" if 
	    ($value and $value != 1 and !(defined $special and ref $special));
    }
#   croak "$kb_fontname does not specify an available font"
#	    unless defined $fontdesc->is_valid;
#   my $kb_redone = $fontdesc->kb_fontname;
#   if ($kb_fontname ne $kb_redone) {
#	warn "Extraneousness in $kb_fontname, trimming to $kb_redone";
#	return newFromKB($class, $kb_redone);
#   }
    return $fontdesc;
}

1;
