package Font::TeXMetrics;
require 5.00404;
use strict;

# Font::TeXMetrics -- manipulate font metrics for TeX (and make TFMs and VFs)
# Copyright (C) 1997-2001  Melissa O'Neill
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 02111-1307, USA.


BEGIN {
    use Exporter   ();
    use vars       qw($VERSION @ISA @EXPORT @EXPORT_OK);
    $VERSION     = 0.01;
    @ISA         = qw(Exporter);
    @EXPORT      = qw();
    @EXPORT_OK   = qw();
}

use Carp;
use Math::Trig;
use Font::Encoding;

#
# $fontmetrics->{ GlyphInfo => { ...<all sorts of stuff about each glyph... },
#		  <Key>     => <Value> }
#

my $defaultKernThreshold = 0;
my $defaultMimicKernThreshold = 0;

sub setDefaultKernThreshold ($) {
    (@_ != 2 or ref($_[0])) and croak 'usage: Font::TeXMetrics->setDefaultKernThreshold( THRESHOLD );';
    $defaultKernThreshold = $_[1];
}

sub setDefaultMimicKernThreshold ($) {
    (@_ != 2 or ref($_[0])) and croak 'usage: Font::TeXMetrics->setDefaultMimicKernThreshold( THRESHOLD );';
    $defaultMimicKernThreshold = $_[1];
}


sub _roundoff ($) {
    my ($number) = @_;
   
    $number = sprintf("%.6f", $number);
    $number =~ s/\.?0+$//;
    return $number;
}

sub _prefixNumber ($) {
    my ($number) = @_;
    
    if ($number >= 0 and $number < 256 and int($number) == $number) {
	return "D $number";
    } else {
        return "R " . _roundoff($number);
    }
}

sub _prefixDesignUnit ($$) {
    my ($number, $units) = @_;
    
    if (defined $units and $units != 1000) {
	_prefixNumber($number / 1000 * $units);
    } else {
	_prefixNumber($number);
    }
}

sub _prefixCharcode ($) {
    my ($char) = @_;
    
    if ($char > 32 and $char < 127 and $char != 40 and $char != 41) {
	return "C " . chr($char);
    } elsif ($char >= 0 and $char < 256 and int($char) == $char) {
        return "D $char";
    } else {
        croak "$char is not a valid character code";
    }
}


sub _cloneGlyphMetrics ($;$) {
    my ($origmetrics,$scale) = @_;
    $scale = 1 unless defined $scale;

    my $newmetrics = {};
    $newmetrics->{Width} = $origmetrics->{Width} * $scale;
    $newmetrics->{BBox}  = [ map { $_ * $scale } @{$origmetrics->{BBox}} ];
    if (defined $origmetrics->{Ligatures}) {
	$newmetrics->{Ligatures}  = { %{$origmetrics->{Ligatures}} };
    }
    my $origkerns = $origmetrics->{Kerns};
    if (defined $origkerns) {
	my $newkerns = {};
	$newmetrics->{Kerns} = $newkerns;
	while ( my ($kernglyph,$kern) = each %$origkerns ) {
	    $newkerns->{$kernglyph} = $kern * $scale;
	}
    }
	
    return $newmetrics;
}


sub clone ($;$) {
    @_ == 1 or @_ == 2 or croak 'usage: $fm2 = $fm1->clone( [ SCALE ] );';
    my ($orig, $scale) = @_;
    $scale = 1 unless defined $scale;
    my $class = ref($orig);
    my $clone = {};

    # Copy simple entries that don't need to be scaled
    foreach my $key ( qw(Comment DesignSize EncodingScheme FamilyName FontID
    		      FontName FullName GlyphInfo IsFixedPitch ItalicAngle
		      KernThreshold MimicKernThreshold Notice Quad Slant Space
		      StdHW StdVW UniqueID Version Weight) ) {
        $clone->{$key} = $orig->{$key} if exists $orig->{$key};
    }

    # Copy simple entries that do need to be scaled
    foreach my $key ( qw(Ascender CapHeight Descender StdHW StdVW
    		      UnderlinePosition UnderlineThickness UniqueID
		      Version XHeight) ) {
        $clone->{$key} = $orig->{$key} * $scale if defined $orig->{$key};
    }

    # Copy object entries that don't need to be scaled
    $clone->{Encoding} = $orig->{Encoding}->clone();
 
    # Handle scaling of spacing; something of a hack...
    $clone->{ExtraSpace} = $orig->{ExtraSpace} * $scale;
    $clone->{Space} = [ map { $_ * $scale } @{$orig->{Space}} ];

    # Scale the FontBBox
    $clone->{FontBBox} = [ map { $_ * $scale } @{$orig->{FontBBox}} ];

    # Copy information for all the glyphs, scaling as necessary...
    my $originfo = $orig->{GlyphInfo};
    if (defined $originfo) {
        my $newinfo = {};
        $clone->{GlyphInfo} = $newinfo;
	while ( my ($glyph,$origmetrics) = each %$originfo ) {
	    $newinfo->{$glyph} = _cloneGlyphMetrics($origmetrics, $scale);
	}
    }
    return bless $clone, $class;
}

sub _clone ($);
sub _clone ($) { 
  my ($value) = @_;
  
  my $type = ref $value;
  
  if ($type) {
      my $clone;
      my ($realpack, $realtype, $id) = ("$value" =~ /^(?:(.*)\=)?([^=]*)\(([^(]*)\)$/);
      if ($realtype eq 'SCALAR') {
          $clone = \(_clone($$value));
      } elsif ($realtype eq 'ARRAY') {
          $clone = [map {_clone($_)} @$value];
      } elsif ($realtype eq 'HASH') {
          $clone = {map {_clone($_)} %$value};
      } else {
          croak "Unsupported type $realpack::$realtype";
      }
      bless $clone, $realpack if defined $realpack;
      return $clone;
  } else {
      return $value;
  }
}

sub new ($) {
    @_ == 1 or croak 'usage: $fm = Font::TeXMetrics->new();';
    my ($thing) = @_;
    my $class = ref($thing) || $thing;
    my $metrics = { GlyphInfo => {} };
    return bless $metrics, $class;
}

sub virtualCopy ($;$) {
    @_ == 1 or @_ == 2 or croak 'usage: $vfm = $fm->virtualCopy( [ SCALE ] );';
    my ($source, $scale) = @_;
    
    my $dest;
    unless (defined $source->{MapInfo}) {
	# First, we copy all the information *except the glyphs* from the
	# base font.
	my $savedglyphinfo = $source->{GlyphInfo};
	delete $source->{GlyphInfo};
	$dest = clone $source;
	$source->{GlyphInfo} = $savedglyphinfo;
    
	$dest->{GlyphInfo} = {};
	$dest->{FontMap} = {};	# FontID => Index
	$dest->{MapInfo} = [];	# Index  => MAPFONT information
	$scale = 1 unless defined $scale;
	$dest->_addFont($source, $scale)->addGlyphsFrom($source);
    } else {
        croak "Can't specify a scale with virtualCopy on virtual font"
	   if (defined $scale and $scale != 1);
        $dest = _clone $source;
    }
    return $dest;
}

sub _addFont ($$;$) {
    @_ == 2 or @_ == 3 or croak 'usage: $vfm->addFont( METRICS [ , SCALE ] );';
    my ($dest, $source, $scale) = @_;
    $scale = 1 unless defined $scale;
    my $mapinfo = $dest->{MapInfo};
    my $fontpos = scalar(@$mapinfo);
    my $fontid  = $source->fontId;
    push @$mapinfo, [$fontid, $source->designsize, 1000 * $scale];
    $dest->{FontMap}->{$fontid} = $fontpos;
    return $dest;
}

sub addGlyphsFrom ($$;@) {
    @_ >= 2 or croak 'usage: $vfm->addGlyphsFrom( METRICS [ , GLYPH ... ] );';
    my ($dest, $source, @glyphs) = @_;
    my $fontid  = $source->fontId;
    my $fontpos = $dest->{FontMap}->{$fontid};
    unless (defined $fontpos) {
        _addFont ($dest, $source);
	$fontpos = $dest->{FontMap}->{$fontid};
    }
    my $scale = $dest->{MapInfo}[$fontpos][2] / 1000;
    
    my $sourceenc = $source->encoding;
    @glyphs = $sourceenc->glyphs unless @glyphs;
    my $source_glyphinfo = $source->{GlyphInfo};
    my $dest_glyphinfo	 =   $dest->{GlyphInfo};
    foreach my $glyph (@glyphs) {
#	print STDERR "$glyph";
	next if $dest->hasGlyph($glyph);
#	print STDERR ".";
	next unless $source->hasGlyph($glyph);
#	print STDERR ".";
	my $glyphinfo = _cloneGlyphMetrics $source_glyphinfo->{$glyph}, $scale;
	my $charcode  = ($sourceenc->charcodesFor($glyph))[0];
	next unless defined $charcode;
#	print STDERR ".";
	my $map = [];
	push (@$map, "(SELECTFONT D $fontpos) (COMMENT $fontid)") if $fontpos;
	my $p_charcode = _prefixCharcode($charcode);
	push (@$map, "(SETCHAR $p_charcode) (COMMENT $glyph)");
	$glyphinfo->{Map} = $map;
	$dest_glyphinfo->{$glyph} = $glyphinfo;
    }
    return $dest;
}

sub stealLigKernsFrom ($$;@) {
    @_ >= 2 or croak 'usage: $vfm->stealKernsFrom( METRICS [ , GLYPH ... ] );';
    my ($dest, $source, @glyphs) = @_;
    my $fontid  = $source->fontId;
    my $fontpos = $dest->{FontMap}->{$fontid};
    my $scale;
    if (defined $fontpos) {
	$scale = $dest->{MapInfo}[$fontpos][2] / 1000;
    } else {
        $scale = 1;
    }
    
    @glyphs = $source->encoding->glyphs unless @glyphs;
    my $source_glyphinfo = $source->{GlyphInfo};
    my $dest_glyphinfo	 =   $dest->{GlyphInfo};
    foreach my $glyph (@glyphs) {
	next unless $source->hasGlyph($glyph);
	my $gmetrics = _cloneGlyphMetrics $source_glyphinfo->{$glyph}, $scale;
	my $dest_gmetrics = $dest_glyphinfo->{$glyph};
	foreach my $ligkern (qw(Ligatures Kerns)) {
	    my $cloned_hash = $gmetrics->{$ligkern};
	    next unless defined $cloned_hash and %$cloned_hash;
	    my $orig_array = $dest_gmetrics->{$ligkern};
	    if (defined $orig_array and @$orig_array) {
		warn "    Glyph $glyph already has ligatures, skipping";
	        next;
	    }
	    $dest_gmetrics->{$ligkern} = $cloned_hash;
	}
    }
    return $dest;
}

sub spaceForMath ($@) {
    @_ >= 2 or croak 'usage: $vfm->spaceForMath( GLYPH [, GLYPH ...] );';
    my ($metrics, @glyphs) = @_;
    
    foreach my $glyph (@glyphs) {
        unless ($metrics->hasGlyph($glyph)) {
	    carp "Can't math-space non-existant glyph '$glyph'";
	    next;
	}
	my ($llx, $lly, $urx, $ury) = $metrics->bbox($glyph);
	my $width = $metrics->width($glyph);
	$width = $urx if $urx > $width;
	#
	# Unfinished code!
	#
	#
	#	?????
	#
	#
    }
    return $metrics;
}

sub newFromAfmFile ($$) {
    @_ == 2 or croak 'usage: $enc = Font::TeXMetrics->newFromAfmFile( FILENAME );';
    my ($thing, $afmfilename) = @_;

    my $metrics = { GlyphInfo => {} };
    bless $metrics, (ref($thing) || $thing);

    my $enc = Font::Encoding->new;

    open (AFMFILE, "$afmfilename") or croak "Can't open AFM file '$afmfilename'";
    
    init: while (<AFMFILE>) {
      next if m/^\s+/;
      last if m/^StartFontMetrics/;
      croak "'StartFontMetrics' expected";
    }
    
    header: while (<AFMFILE>) {
	chomp;
	tr/\cM//d;	  # Kill any CRs (in case we have CR/LF as line end)
	next if m/^\s+$/; # Skip blank lines...
	s/^Comment (?=(?:UniqueID|FontID|DesignSize|CharacterCodingScheme|Space|ExtraSpace|Quad))//;
	my ($key, $value) = split(' ', $_, 2);
	
	last if $key eq 'StartCharMetrics';
	$metrics->{$key} = $value;
    }
    
    charmetrics: while (<AFMFILE>) {
	chomp;
	tr/\cM//d;	  # Kill any CRs (in case we have CR/LF as line end)
	next if m/^\s+$/; # Skip blank lines...
	last if m/EndCharMetrics/;
	my $charcode;
	my $glyphname;
	my $glyphinfo = {};
	foreach my $keypair (split(/\s+;\s*/, $_)) {
	    my ($key, @values) = split(' ', $keypair, -1);
	    if ($key eq 'C') {
		$charcode = $values[0];
	    } elsif ($key eq 'WX' or $key eq 'W0X') {
		$glyphinfo->{Width} = $values[0];
	    } elsif ($key eq 'N') {
		$glyphname = $values[0];
	    } elsif ($key eq 'B') {
	    	$glyphinfo->{BBox} = [ @values ];
	    } elsif ($key eq 'L') {
		$glyphinfo->{Ligatures}->{$values[0]} = $values[1];
	    } else {
		warn "Unspported char metric data (ignored)";
	    }
	}
	unless (defined $glyphname) {
	    carp "Metric file does not name glyph at $charcode, calling it glyph$charcode";
	    $glyphname = "glyph$charcode";
	}
	$enc->encodeAs($charcode, $glyphname) if $charcode >= 0;
	$metrics->{GlyphInfo}->{$glyphname} = $glyphinfo;
    }
    
    my $foundKernPairs = 0;
    
    while (<AFMFILE>) {
	if (m/^StartKernPairs/) {
	  $foundKernPairs = 1;
	  last;
	}
    }
    
    if ($foundKernPairs) {
	while(<AFMFILE>) {
	    last if m/^EndKernPairs/;
	    next if m/^\s+$/;
	    chomp;
	    tr/\cM//d;	  # Kill any CRs (in case we have CR/LF as line end)
	    my ($kerntype, @values) = split(' ', $_, -1);
	    if ($kerntype eq 'KPX') {
	      my ($leftglyph, $rightglyph, $kern) = @values;
	      $metrics->{GlyphInfo}->{$leftglyph}->{Kerns}->{$rightglyph} = $kern;
	    } else {
	      warn "Unsupported kern type '$kerntype'";
	    }
	}
    }
    close AFMFILE;

    ## Check to make sure we have definitions for required keywords
    foreach my $key (qw(FontName FontBBox)) {
        croak "AFM file does not define $key" unless defined $metrics->{$key};
    }

    ## Convert a few Keyword values into Perl values

    # Convert IsFixedPitch from string to boolean
    $metrics->{IsFixedPitch} = ((defined $metrics->{IsFixedPitch} and $metrics->{IsFixedPitch} eq 'true') or defined $metrics->{CharWidth});

    # Convert BBox from string to array
    $metrics->{FontBBox} = [ split(' ', $metrics->{FontBBox}, 4) ];

    ## Set basic font info to default values for optional keyword...

    # Default for EncodingScheme
    $metrics->{EncodingScheme} = 'FontSpecific'
    	unless defined $metrics->{EncodingScheme};

    # Default for Weight
    $metrics->{Weight} = 'Regular'
    	unless defined $metrics->{Weight};

    # Default for ItalicAngle
    $metrics->{ItalicAngle} = 0
    	unless defined $metrics->{ItalicAngle};

    # Defaults for CapHeight, XHeight, Ascender, and Descender
    my @heightglyphs  = qw(H x d p);
    my @heightindexes = (3, 3, 3, 1);			# ury => 3, lly => 1
    my @sillyguesses  = (750, 500, 750, 250);
    foreach my $heightword (qw(CapHeight XHeight Ascender Descender)) {
	my $heightglyph = shift @heightglyphs;
	my $heightindex = shift @heightindexes;
	my $sillyguess = shift @sillyguesses;
	unless (defined $metrics->{$heightword}) {
	    if ($metrics->hasGlyph($heightglyph)) {	# Use value from char
		$metrics->{$heightword} = $enc->bbox($heightglyph)->[$heightindex];
	    } else { 					# Use silly value
		$metrics->{$heightword} = $sillyguess; 
	    }
	}
    }

    # Set TeX specific font info to default values if undefined...

    # Set DesignSize if necessary
    if (defined $metrics->{DesignSize}) {
        $metrics->{DesignSize} = (split(' ', $metrics->{DesignSize}, 2))[0];
    } else {
	$metrics->{DesignSize} = 10;
    }

    # Set FontID if necessary
    unless (defined $metrics->{FontID}) {
        my $fontid = $afmfilename;
	$fontid =~ s#\.afm$##;
	$fontid =~ s#^.*/##;
	$metrics->{FontID} = $fontid;
    }

    # Set Slant if necessary    
    $metrics->{Slant} = _roundoff(tan(-deg2rad($metrics->{ItalicAngle})))
    	unless defined $metrics->{Slant};

    # Figure out the width of an interword space, used below    
    my $spacewidth;
    if ($metrics->hasGlyph('space')) {
	$spacewidth = $metrics->width('space');
    } elsif ($metrics->hasGlyph('m')) {	# Use value from char
	$spacewidth = $metrics->width('m') / 4;
    } else { 					# Use silly value
	$spacewidth = 400; 
    }

    # Set Space if necessary   
    if (defined $metrics->{Space}) {
	$metrics->{Space} = [ split(' ', $metrics->{Space}, 3) ];
    } else {
        unless ($metrics->{IsFixedPitch}) {
	    $metrics->{Space} = [ _roundoff($spacewidth),
				  _roundoff($spacewidth * 3 / 5),
				  _roundoff($spacewidth / 4.2) ];
	    # Note, if we were using the CMR conventions, we'd use these
	    # more generous proportions.
	    # $metrics->{Space} = [ _roundoff($spacewidth),
	    #		            _roundoff($spacewidth / 2),
	    #		            _roundoff($spacewidth / 3) ];
	} else {
	    $metrics->{Space} = [ _roundoff($spacewidth), 0, 0 ];
	}
    }
    
    # Set ExtraSpace if necessary   
    unless (defined $metrics->{ExtraSpace}) {
        unless ($metrics->{IsFixedPitch}) {
	    $metrics->{ExtraSpace} = $metrics->{Space}[2];
	} else {
	    $metrics->{ExtraSpace} = $spacewidth;
	}
    }

    # Set Quad if necessary   
    $metrics->{Quad} = 1000 unless (defined $metrics->{Quad});

    # Set Kerning Threshold   
    $metrics->{KernThreshold} = $defaultKernThreshold unless (defined $metrics->{KernThreshold});

    # Set Mimic Kerning Threshold
    $metrics->{MimicKernThreshold} = $defaultMimicKernThreshold unless (defined $metrics->{MimicKernThreshold});

    # Try to be smart if we have a 'FontSpecific' encoding, maybe the
    # real name is in CharacterCodingScheme.
    
    my $encname = $metrics->{EncodingScheme};
    if ($encname eq 'FontSpecific'
	    and exists $metrics->{CharacterCodingScheme}) {
	$encname = $metrics->{CharacterCodingScheme};
	$encname =~ s/[ \t]/-/g;
    }
    $enc->renameTo($encname);
        
    $metrics->{Encoding} = $enc;

    return $metrics;
}

# These methods simply read data out of fields in the $font data structure

sub encoding ($) {
    @_ == 1 or croak 'usage: $enc = $font->encoding();';
    return $_[0]->{Encoding};
}

sub xHeight ($) {
    @_ == 1 or croak 'usage: $font->isfixedpitch();';
    return $_[0]->{XHeight};
}

sub fontId ($) {
    @_ == 1 or croak 'usage: $font->fontID();';
    return $_[0]->{FontID};
}

sub fontName ($) {
    @_ == 1 or croak 'usage: $font->fontName();';
    return $_[0]->{FontName};
}

sub kernThreshold ($) {
    @_ == 1 or croak 'usage: $font->kernThreshold();';
    return $_[0]->{KernThreshold};
}

sub mimicKernThreshold ($) {
    @_ == 1 or croak 'usage: $font->mimicKernThreshold();';
    return $_[0]->{MimicKernThreshold};
}

sub family ($) {
    @_ == 1 or croak 'usage: $font->family();';
    return $_[0]->{FamilyName};
}

sub isFixedPitch ($) {
    @_ == 1 or croak 'usage: $font->isfixedpitch();';
    return $_[0]->{IsFixedPitch};
}

sub slant ($) {
    @_ == 1 or croak 'usage: $slant = $font->slant();';
    return $_[0]->{Slant};
}

sub spacing ($) {
    @_ == 1 or croak 'usage: ($space, $stretch, $shrink) = $font->spacing();';
    return @{$_[0]->{Space}};
}

sub extraspace ($) {
    @_ == 1 or croak 'usage: $extraspace = $font->extraspace();';
    return $_[0]->{ExtraSpace};
}

sub designsize ($) {
    @_ == 1 or croak 'usage: $designsize = $font->designsize();';
    return $_[0]->{DesignSize};
}

sub quad ($) {
    @_ == 1 or croak 'usage: $quad = $font->quad();';
    return $_[0]->{Quad};
}

# These methods read data out of fields in the font metrics data structure,
# but need to do some kind of calculation

sub isItalic ($) {
    @_ == 1 or croak 'usage: ... if $font->isItalic();';
    my ($metrics) = @_;
    return $metrics->{ItalicAngle} != 0;
}

sub glyphs ($) {
    @_ == 1 or croak 'usage: @glyphs = $font->glyphs();';
    my ($metrics) = @_;

    return keys %{$metrics->{GlyphInfo}};
}

# These methods alter data in the font metrics data structure.

sub setEncodingTo ($$) {
    @_ == 2 or croak 'usage: $font->setEncodingTo( ENCODING );';
    my ($metrics, $enc) = @_;
    $metrics->{Encoding} = $enc;
    return $metrics;
}

sub setFontIdTo ($$) {
    @_ == 2 or croak 'usage: $font->setFontIdTo( FONTNAME );';
    my ($metrics, $name) = @_;
    $metrics->{FontID} = $name;
    return $metrics;
}

sub setKernThresholdTo ($$) {
    @_ == 2 or croak 'usage: $font->setKernThresholdTo( UNITS );';
    my ($metrics, $kernThreshold) = @_;
    $metrics->{KernThreshold} = $kernThreshold;
    return $metrics;
}

sub setMimicKernThresholdTo ($$) {
    @_ == 2 or croak 'usage: $font->setKernThresholdTo( UNITS );';
    my ($metrics, $kernThreshold) = @_;
    $metrics->{MimicKernThresholdTo} = $kernThreshold;
    return $metrics;
}

sub hasGlyph {
    @_ == 2 or croak 'usage: ... if ($font->hasGlyph( GLYPH );';
    my ($metrics, $glyph) = @_;
    return defined $metrics->{GlyphInfo}->{$glyph};
}

sub _deleteFromHash {
    my $hash = shift;
    foreach my $key (@_) {
	delete $hash->{$key}
    }
}

sub deleteGlyphs {
    @_ >= 2 or croak 'usage: $font->deleteGlyphs( GLYPH ... );';
    my $metrics = shift;
    my $glyphinfo = $metrics->{GlyphInfo};
    _deleteFromHash $glyphinfo, @_;
    return $metrics;
}

sub deleteGlyphsAndLigKerns {
    @_ >= 2 or croak 'usage: $font->deleteGlyphsAndLigKerns( GLYPH ... );';
    my $metrics = shift;
    my $glyphinfo = $metrics->{GlyphInfo};
    _deleteFromHash $glyphinfo, @_;
    while ( my ($glyph,$gmetrics) = each %$glyphinfo ) {
	my $ligatures = $gmetrics->{Ligatures};
	if (defined $ligatures) {
	    _deleteFromHash $ligatures, @_;
	}
	my $kerns = $gmetrics->{Kerns};
	if (defined $kerns) {
	    _deleteFromHash $kerns, @_;
	}
    }
    return $metrics;
}

sub _renameHash {
    my $hash = shift;
    my %newname = @_;
    foreach my $key (keys %$hash) {
	delete $hash->{$key} unless $newname{$key};
    }
    while (@_) {
        my $data  = delete $hash->{(shift)};
	my $key   = shift;
	$hash->{$key} = $data if defined $data;
    }
}

sub restrictAndRenameGlyphs {
    (@_ >= 3 and (@_ % 2) == 1)
	  or croak 'usage: $font->restrictGlyphs( GLYPH, NEWNAME ... );';
    my $metrics = shift;
    my $newenc = $metrics->encoding->clone->restrictAndRenameGlyphs(@_);
    $metrics->setEncodingTo($newenc);
    my $glyphinfo = $metrics->{GlyphInfo};
    _renameHash $glyphinfo, @_;
    while ( my ($glyph,$gmetrics) = each %$glyphinfo ) {
	my $ligatures = $gmetrics->{Ligatures};
	if (defined $ligatures) {
	    _renameHash $ligatures, @_;
	}
	my $kerns = $gmetrics->{Kerns};
	if (defined $kerns) {
	    _renameHash $kerns, @_;
	}
    }

    return $metrics;
}

sub addSmallCapsKerns {
    @_ == 1 or croak 'usage: $font->addSmallCapsKerns();';
    my ($metrics) = @_;
    my $glyphinfo = $metrics->{GlyphInfo};
    foreach my $leftkern ('A'..'Z') {
	my $kerns = $glyphinfo->{$leftkern}{Kerns};
	next unless defined $kerns;
        my $sc_leftkern = "${leftkern}small";
#	print STDERR "$leftkern/$sc_leftkern vs\n";
	my $sc_kerns = $glyphinfo->{$sc_leftkern}{Kerns};
	croak "Crazy! No $sc_leftkern" unless defined $sc_kerns;
	foreach my $rightkern ('A'..'Z') {
	    my $kern     = $kerns->{$rightkern};
	    next unless defined $kern;
	    my $sc_rightkern = "${rightkern}small";
#	    print STDERR "\t$rightkern/$sc_rightkern\n";
	    my $sc_kern  = $sc_kerns->{$sc_rightkern};
	    $sc_kern = 0 unless defined $sc_kern;
#	    my $avg_kern = ($kern + $sc_kern) / 2;
#	    my $avg_kern = $sc_kern;
	    my $avg_kern = ($kern + $sc_kern) / 3;
	    $kerns->{$sc_rightkern} = $avg_kern;
	    $sc_kerns->{$rightkern} = $avg_kern;
	}
    }
}

sub _ensureGlyphExists {
    unless (hasGlyph @_) {
        my ($metrics, $glyph) = @_;
#	croak "Information requested on non-existent glyph '$glyph'"
	confess "Information requested on non-existent glyph '$glyph'"
    }
}

sub bbox ($;$) {
    if (@_ == 2) {
	_ensureGlyphExists @_;
	my ($metrics, $glyph) = @_;
	if (defined (my $bbox = $metrics->{GlyphInfo}->{$glyph}->{BBox})) {
	    return @$bbox;
	} else {
	    carp "No bounding box defined for '$glyph', inventing one.";
	    return (0, 0, 0, 0); # Maybe we can do better than this...
	}
    } elsif (@_ == 1) {
	my ($metrics) = @_;
	return @{$metrics->{FontBBox}};
    } else {
        croak 'usage: ($llx, $lly, $urx, $ury) = $font->bbox( GLYPH );\nor: ($llx, $lly, $urx, $ury) = $font->bbox();';
    }
}

sub setBboxTo ($$@) {
    @_ == 6 or croak 'usage: $font->setBboxTo( GLYPH, LLX, LLY, URX, URY );';
    my ($metrics, $glyph, @bbox) = @_;
    _ensureGlyphExists ($metrics, $glyph);
    $metrics->{GlyphInfo}->{$glyph}->{BBox} = [@bbox];
    return $metrics;
}

sub width {
    @_ == 2 or croak 'usage: $width = $font->width( GLYPH );';
    _ensureGlyphExists @_;
    my ($metrics, $glyph) = @_;
    return $metrics->{GlyphInfo}->{$glyph}->{Width};
}

sub setWidthTo {
    @_ == 3 or croak 'usage: $font->setWidthTo( GLYPH, WIDTH );';
    my ($metrics, $glyph, $width) = @_;
    _ensureGlyphExists ($metrics, $glyph);
    $metrics->{GlyphInfo}->{$glyph}->{Width} = $width;
    return $metrics;
}

sub widen {
    @_ == 3 or croak 'usage: $font->setWidthTo( GLYPH, DELTA );';
    my ($metrics, $glyph, $delta) = @_;
    _ensureGlyphExists ($metrics, $glyph);
    my $glyphinfo = $metrics->{GlyphInfo}->{$glyph};
    $glyphinfo->{Width} += $delta;
    if (defined (my $map = $glyphinfo->{Map})) {
        my $halfdelta = $delta / 2;
	my $closure = sub {
	    my ($units) = @_;
	    my $p_halfdelta = _prefixDesignUnit($halfdelta, $units);
	    return "(MOVERIGHT $p_halfdelta) (COMMENT $delta tracking)";
	};
	unshift @$map, $closure;
    }
    return $metrics;
}


#sub hasKernsFor ($$) {
#    @_ == 2 or croak 'usage: ... if $font->hasKernsFor( GLYPH )};';
#    _ensureGlyphExists @_;
#    my ($metrics, $glyph) = @_;
#    my $kerns = $metrics->{GlyphInfo}->{$glyph}->{Kerns};
#    return (defined $kerns and scalar(%$kerns));
#}
#
#sub hasLigaturesFor ($$) {
#    @_ == 2 or croak 'usage: ... if $font->hasLigaturesFor( GLYPH )};';
#    _ensureGlyphExists @_;
#    my ($metrics, $glyph) = @_;
#    my $ligatures = $metrics->{GlyphInfo}->{$glyph}->{Ligatures};
#    return (defined $ligatures and scalar(%$ligatures));
#}

sub kernsFor ($$) {
    @_ == 2 or croak 'usage: %kerns = %{$font->kernsFor( GLYPH )};';
    _ensureGlyphExists @_;
    my ($metrics, $glyph) = @_;
    return $metrics->{GlyphInfo}->{$glyph}->{Kerns};
}

sub ligaturesFor ($$) {
    @_ == 2 or croak 'usage: %kerns = %{$font->ligaturesFor( GLYPH )};';
    _ensureGlyphExists @_;
    my ($metrics, $glyph) = @_;
    return $metrics->{GlyphInfo}->{$glyph}->{Ligatures};
}

sub _forceLigaturesFor ($$) {
    @_ == 2 or croak 'usage: %kerns = %{$font->_forceLigaturesFor( GLYPH )};';
    _ensureGlyphExists @_;
    my ($metrics, $glyph) = @_;
    return \%{$metrics->{GlyphInfo}->{$glyph}->{Ligatures}};
}


sub charwd ($$) {
    @_ == 2 or croak 'usage: $charwd = $font->charwd( GLYPH );';
    return width(@_);
}

sub charht ($$) {
    @_ == 2 or croak 'usage: $charht = $font->charht( GLYPH );';
    _ensureGlyphExists @_;
    my ($metrics, $glyph) = @_;
    my ($llx, $lly, $urx, $ury) = $metrics->bbox($glyph);
    my $xheight = $metrics->xHeight;

    # We might expect charht to just be $ury, but that would cause TeX
    # to misplace accents on characters whose height ought to be seen
    # as just xheight. We use the same algorithm that Tom Rokicki's
    # afm2tfm uses; were it not for freely available source, I'd not
    # have been able to code that algorithm here.

    return $ury unless $xheight > 50;
    
    foreach my $accent (qw(acute tilde caron dieresis)) {
	if ($metrics->hasGlyph($glyph . $accent)
		and $metrics->hasGlyph($accent)) {
	    my ($agllx, $aglly, $agurx, $agury) = $metrics->bbox($glyph . $accent);
	    my ($allx, $ally, $aurx, $aury) = $metrics->bbox($accent);
	    return _roundoff($agury - $aury + $xheight);
	}
    }   
    
    return $ury;
}

sub chardp ($$) {
    @_ == 2 or croak 'usage: $chardp = $font->chardp( GLYPH );';
    _ensureGlyphExists @_;
    my ($metrics, $glyph) = @_;
    my ($llx, $lly, $urx, $ury) = $metrics->bbox($glyph);
    return _roundoff(0 - $lly);
}

sub charic ($$) {
    @_ == 2 or croak 'usage: $font->charic( GLYPH );';
    _ensureGlyphExists @_;
    my ($metrics, $glyph) = @_;
    my ($width) = $metrics->width($glyph);
    my ($llx, $lly, $urx, $ury) = $metrics->bbox($glyph);
    my $ic = _roundoff($urx - $width);
    return $ic > 0 ? $ic : 0;
}

sub _fetchlist($\%) {
    my ($key, $hashref) = @_;
    
    my $list = $hashref->{$key};
    return @{\@{$list}};
}

# The hashes below detail compound letters that kern similar to base
# letters. The data is taken from both afm2tfm and fontinst.

my %_kernLike = (
    'A' => [qw(Acircumflex Aogonek Aring Abreve Agrave Adieresis Atilde Aacute)],
    'C' => [qw(Cacute Ccedilla Ccaron)],
    'D' => [qw(Dcaron)],
    'E' => [qw(Ecircumflex Eogonek Ecaron Egrave Edieresis Eacute)],
    'Eth' => [qw(Dbar)],
    'G' => [qw(Gbreve)],
    'I' => [qw(Icircumflex Igrave Idieresis Iacute Idotaccent)],
    'L' => [qw(Lslash Lacute)],
    'N' => [qw(Nacute Ntilde Ncaron)],
    'O' => [qw(Ograve Otilde Oacute Ocircumflex Odieresis Ohungarumlaut Oslash)],
    'R' => [qw(Racute Rcaron)],
    'S' => [qw(Scedilla SS Scaron Sacute)],
    'T' => [qw(Tcaron Tcedilla)],
    'U' => [qw(Ucircumflex Uring Ugrave Udieresis Uhungarumlaut Uacute)],
    'Y' => [qw(Ydieresis Yacute)],
    'Z' => [qw(Zacute Zdotaccent Zcaron)],
    'a' => [qw(acircumflex aogonek aring agrave adieresis aacute atilde)],
    'c' => [qw(ccedilla)],
    'e' => [qw(ecircumflex eogonek egrave edieresis eacute)],
    'f' => [qw(ff longs longss)],
    'i' => [qw(icircumflex igrave idieresis iacute)],
    'n' => [qw(ntilde)],
    'o' => [qw(oslash ograve oacute ocircumflex otilde odieresis)],
    's' => [qw(scedilla scaron)],
    't' => [qw(tcedilla)],
    'u' => [qw(uacute ucircumflex ugrave udieresis)],
    'y' => [qw(yacute ydieresis)],
    'z' => [qw(zcaron)]
);

my %_kernLeftLike = (
  'J' => [qw(IJ Ng)],
  'e' => [qw(oe)],
  'i' => [qw(ffi fi)],
  'j' => [qw(ij ng)],
  'l' => [qw(fl ffl)],
  'p' => [qw(sp)],
  't' => [qw(ct st)]
);

my %_kernRightLike = (
  'I' => [qw(IJ)],
  'L' => [qw(Lcaron)],
  'N' => [qw(Ng)],
  'c' => [qw(ct)],
  'f' => [qw(fl ffi ffl fi)],
  'i' => [qw(ij)],
  'n' => [qw(ng)],
  'o' => [qw(oe)],
  's' => [qw(sp st)]
);

# Mabe these two functions should pull information from the font, but
# right now they take it from a static table, and then check it a little
# with the font.

sub similarLeftKerns ($$) { 
    @_ == 2 or croak 'usage: @leftkerns = $font->similarLeftKerns( GLYPH );';
    _ensureGlyphExists @_;
    my ($metrics, $glyph) = @_;
    my @kerns = (_fetchlist ($glyph, %_kernLike), _fetchlist ($glyph, %_kernLeftLike));
    return grep {$metrics->hasGlyph($_)} @kerns;
}

sub similarRightKerns ($$) {
    @_ == 2 or croak 'usage: @leftkerns = $font->similarRightKerns( GLYPH );';
    _ensureGlyphExists @_;
    my ($metrics, $glyph) = @_;
    my @kerns = (_fetchlist ($glyph, %_kernLike), _fetchlist ($glyph, %_kernRightLike));
    return grep {$metrics->hasGlyph($_)} @kerns;
}

my $_texLigatures = {
  'f'		=> {'i' => 'fi', 'l' => 'fl', 'f' => 'ff'},
# 'space'	=> {'l' => 'lslash', 'L' => 'Lslash'},
# 'exclam'	=> {'quoteleft' => 'exclamdown'},
# 'question'	=> {'quoteleft' => 'questiondown'},
  'hyphen'	=> {'hyphen' => 'endash'},
  'endash'	=> {'hyphen' => 'emdash'},
  'quoteleft'   => {'quoteleft' => 'quotedblleft'},
  'quoteright'	=> {'quoteright' => 'quotedblright'}
};

my $_expertLigatures = {
  'f'		=> {'i' => 'fi', 'l' => 'fl', 'f' => 'ff', 'fl' => 'ffl', 'fi' => 'ffi'},
  'ff'		=> {'i' => 'ffi', 'l' => 'ffl'},
# 'endash'	=> {'hyphen' => 'threequartersemdash'},
#   ^--- I put this in, but eventually decided that I hated it
};

sub addLigatures ($$) {
    my ($metrics, $ligatureset) = @_;
    @_ == 2 or croak 'usage: $font->addLigatures( LIGATURESET );';
    
    while (my ($leftglyph,$extraligatures) = each %$ligatureset) {
	next unless $metrics->hasGlyph($leftglyph);
	my $ligatures = $metrics->_forceLigaturesFor($leftglyph);
	while (my ($rightglyph,$bothglyph) = each %{$extraligatures}) {
	    $ligatures->{$rightglyph} = $bothglyph;
	    my $kerns = $metrics->kernsFor($leftglyph);
	    if (defined $kerns) {
		delete $kerns->{$rightglyph};
	    }
	}
    }
    return $metrics;
}

sub addTexLigatures ($) {
    my ($metrics) = @_;
    @_ == 1 or croak 'usage: $font->addTexLigatures();';

    $metrics->addLigatures($_texLigatures);
    return $metrics;
}

sub addExpertLigatures ($) {
    my ($metrics) = @_;
    @_ == 1 or croak 'usage: $font->addExpertLigatures();';

    $metrics->addLigatures($_expertLigatures);
    return $metrics;
}

sub deleteLigaturesFor ($$) {
    @_ == 2 or croak 'usage: %kerns = %{$font->deleteLigaturesFor( GLYPH )};';
    _ensureGlyphExists @_;
    my ($metrics, $glyph) = @_;
    delete $metrics->{GlyphInfo}->{$glyph}->{Ligatures};
    return $metrics;
}

sub _kernLines ($$$) {
    my ($metrics, $leftglyph, $units) = @_;
    my $kerns = $metrics->kernsFor($leftglyph);
    my $threshold = $metrics->kernThreshold;
    my $mthreshold = $metrics->mimicKernThreshold;
    return qw() unless defined $kerns;

    my @kernlines = ();
    my $enc = $metrics->encoding;
    while (my ($rightglyph,$kern) = each %$kerns) {
	my $p_kern      = _prefixDesignUnit($kern, $units);
    	next unless abs($kern) > $threshold;
	foreach my $rightcode ($enc->charcodesFor($rightglyph)) {
	    my $p_rightcode = _prefixCharcode($rightcode);
	    push @kernlines, "   (KRN $p_rightcode $p_kern) (COMMENT $leftglyph+$rightglyph)";
	}
	next unless abs($kern) > $mthreshold;
	foreach my $likeglyph ($metrics->similarRightKerns($rightglyph)) {
	    unless (defined $kerns->{$likeglyph}) {
		foreach my $likecode ($enc->charcodesFor($likeglyph)) { 
		    my $p_likecode = _prefixCharcode($likecode);
		    push @kernlines, "   (KRN $p_likecode $p_kern) (COMMENT $leftglyph+$likeglyph)";
		}
	    }
	}
    }
    return @kernlines;
}

sub _ligLines ($$) {
    my ($metrics, $leftglyph) = @_;
    my $ligs = $metrics->ligaturesFor($leftglyph);
    return qw() unless defined $ligs;

    my @liglines;
    my $enc = $metrics->encoding;

    while (my ($rightglyph,$bothglyph)= each %$ligs) {
	next unless $metrics->hasGlyph($bothglyph);
	foreach my $rightcode ($enc->charcodesFor($rightglyph)) {
	    my $p_rightcode = _prefixCharcode($rightcode);
	    my @ligcodes = $enc->charcodesFor($bothglyph);
	    warn ("'$bothglyph' is at positions " . join(@ligcodes, ' and ') . " in output encoding, causing confusion for the '$leftglyph' + '$rightglyph' -> '$bothglyph' ligature") if scalar(@ligcodes) > 1;
	    foreach my $ligcode (@ligcodes) {
	        my $p_ligcode = _prefixCharcode($ligcode);
		push @liglines, "   (LIG $p_rightcode $p_ligcode) (COMMENT $leftglyph+$rightglyph -> $bothglyph)";
	    }
	}
    }
    return @liglines;
}


sub writePropertyList ($;$) {
    @_ == 1 or @_ == 2 or croak 'usage: $font->writePropertyList( [ FILENAME ] );';
    my ($metrics, $plfilename) = @_;
    
    my $units = 100;
    my $mapinfo = $metrics->{MapInfo};
    $plfilename = $metrics->fontId . (defined $mapinfo ? '.vpl' : '.pl')
	unless defined $plfilename;
    
    open (PLFILE, ">$plfilename") or croak "Couldn't open '$plfilename'";
    {
	my @spacing 	    = $metrics->spacing;
	my $p_designsize    = _prefixNumber($metrics->designsize);
	my $p_designunits   = _prefixNumber($units);
	my $family 	    = $metrics->family;
	my $codingscheme    = $metrics->encoding->name;
	my $p_slant 	    = _prefixNumber($metrics->slant);
	my $p_space 	    = _prefixDesignUnit($spacing[0], $units);
	my $p_stretch 	    = _prefixDesignUnit($spacing[1], $units);
	my $p_shrink 	    = _prefixDesignUnit($spacing[2], $units);
	my $p_xheight 	    = _prefixDesignUnit($metrics->xHeight, $units);
	my $p_quad 	    = _prefixDesignUnit($metrics->quad, $units);
	my $p_extraspace    = _prefixDesignUnit($metrics->extraspace, $units);
	printf PLFILE <<"END";
(DESIGNSIZE $p_designsize)
(DESIGNUNITS $p_designunits)
(COMMENT DESIGNSIZE (1 em) is in points)
(COMMENT other dimensions are multiples of DESIGNSIZE/$units)
(COMMENT Generated by FontKit, (C) 1998-2000 Melissa O'Neill)
(SEVENBITSAFEFLAG FALSE)
(FAMILY $family)
(CODINGSCHEME $codingscheme)
(FONTDIMEN
   (SLANT $p_slant)
   (SPACE $p_space) 
   (STRETCH $p_stretch) 
   (SHRINK $p_shrink) 
   (XHEIGHT $p_xheight) 
   (QUAD $p_quad) 
   (EXTRASPACE $p_extraspace) 
   )
END
    }
    if (defined $mapinfo) {
	my $fontpos = 0;
        foreach my $entry (@$mapinfo) {
	    my ($fontid, $fontdsize, $fontat) = @$entry;
	    my $p_fontat = _prefixDesignUnit($fontat, $units);
	    print PLFILE <<"END";
(MAPFONT D $fontpos
   (FONTNAME $fontid)
   (FONTAT $p_fontat)
   (FONTDSIZE D $fontdsize)
   )
END
	    ++$fontpos;
	}
    }
    
    my $enc = $metrics->encoding;
    foreach my $charcode (0..255) {
	my $glyph = $enc->glyphFor($charcode);
	next unless defined $glyph;
	unless ($metrics->hasGlyph($glyph)) {
	  warn "\tMissing glyph '$glyph' for " . $metrics->encoding->name . "[$charcode]\n";
	  next;
	}
	my $p_charcode = _prefixCharcode($charcode);
	print PLFILE "(CHARACTER $p_charcode (COMMENT $glyph)\n";
	my @fieldvalues = ($metrics->charwd($glyph), $metrics->charht($glyph),
			   $metrics->chardp($glyph), $metrics->charic($glyph));
	foreach my $field (qw(CHARWD CHARHT CHARDP CHARIC)) {
	    my $fieldvalue = shift @fieldvalues;
	    next unless $fieldvalue;
	    my $p_fieldvalue = _prefixDesignUnit($fieldvalue, $units);
	    print PLFILE "   ($field $p_fieldvalue)\n" if $fieldvalue;
	}
	if (defined (my $map = $metrics->{GlyphInfo}->{$glyph}->{Map})) {
	    print PLFILE "   (MAP\n";
	    foreach my $mapline (@$map) {
	        my $line;
	        if (ref $mapline) {
		    $line = &$mapline($units);
		} else {
		    $line = $mapline;
		}
		print PLFILE "      $line\n";
	    }
	    print PLFILE "      )\n";
	}
	print PLFILE "   )\n";
    }
    
    print PLFILE "(LIGTABLE\n";
    
    my %alreadydone;
    
    foreach my $charcode (0..255) {
	my $leftglyph = $enc->glyphFor($charcode);
	next unless defined $leftglyph;
	next unless $metrics->hasGlyph($leftglyph);
	foreach my $likeglyph ($metrics->similarLeftKerns($leftglyph)) {
	    $alreadydone{$likeglyph} = 1;
	}
    }
    
    foreach my $charcode (0..255) {
	my $leftglyph = $enc->glyphFor($charcode);
	next unless defined $leftglyph;
	next unless $metrics->hasGlyph($leftglyph);
	my @liglines  = $metrics->_ligLines($leftglyph);
	my @kernlines = $metrics->_kernLines($leftglyph, $units);
	next unless @liglines or @kernlines;
	next if $alreadydone{$leftglyph};
	$alreadydone{$leftglyph} = 1;
	my $ligkerncount = 0;
	my @ligkernlines;
	my @labellines = ();
	foreach my $leftcode ($enc->charcodesFor($leftglyph)) {
	    my $p_leftcode = _prefixCharcode($leftcode);
	    push @labellines,  "   (LABEL $p_leftcode) (COMMENT $leftglyph)";
	}
	if (@liglines) {
	    push @ligkernlines, @labellines;
	    push @ligkernlines, @liglines;
	    @labellines = ();
	    $ligkerncount += scalar(@liglines);
	}
	if (@kernlines) {
	    foreach my $likeglyph ($metrics->similarLeftKerns($leftglyph)) {
		$alreadydone{$likeglyph} = 1;
		foreach my $likecode ($enc->charcodesFor($likeglyph)) { 
		    my @otherligkernlines = ($metrics->_ligLines($likeglyph), $metrics->_kernLines($likeglyph));
		    my $p_likecode = _prefixCharcode($likecode);
		    if (@otherligkernlines) {
			push @ligkernlines, '-skip-' if @ligkernlines;
			push @ligkernlines, "   (LABEL $p_likecode) (COMMENT $likeglyph)";
			$ligkerncount += scalar(@otherligkernlines);
			push @ligkernlines, @otherligkernlines;
		    } else {
			push @labellines, "   (LABEL $p_likecode) (COMMENT $likeglyph)";
		    }
		}
	    }
	    foreach my $entry (@ligkernlines) {
#	        print "CHECKING $entry -- $ligkerncount\n";
		--$ligkerncount if ($entry =~ m/\(LIG/ or $entry =~m/\(KRN/);
		if ($entry eq '-skip-') {
		    if ($ligkerncount) {
		        my $p_ligkerncount = _prefixNumber($ligkerncount);
			$entry = "   (SKIP $p_ligkerncount)";
		    } else {
		    	$entry = "   (COMMENT SKIP D 0)";
#		        undef $entry;
		    }
		}
	    }
#	    @ligkernlines = grep {$_} @ligkernlines;
	    push @ligkernlines, @labellines;
	    push @ligkernlines, @kernlines;
	}
	push @ligkernlines, "   (STOP)\n";
	print PLFILE join("\n", @ligkernlines);
    }
    
    print PLFILE "   )\n";      
    print PLFILE "(COMMENT Phew!)\n";
    close PLFILE;
    return $metrics;
}

sub writeTFM ($) {
    (@_ >= 1 and @_ <= 2) or croak 'usage: $font->writeTFM([TFMFILE]);';
    my ($metrics, $tfmfile) = @_;
    my $fontid = $metrics->fontId;
    
    croak "You should be writing a VF file for $fontid" if $metrics->{MapInfo};
    my $plfile = "$fontid.pl";
    $tfmfile   = "$fontid.tfm" unless defined $tfmfile;
    writePropertyList($metrics, $plfile);
    my $cmd = "pltotf $plfile $tfmfile";
    my $diagnostics = `$cmd`;
    croak "$cmd failed" unless defined $diagnostics;
    $diagnostics =~ s/^/\tpltotf: /mg if $diagnostics ne '';
    $diagnostics =~ s/I had to round/rounded/mg;
    print STDERR $diagnostics;
    unlink $plfile unless -e '.keep_plfiles';
    return $metrics;
}

sub writeVF ($) {
    (@_ >= 1 and @_ <= 3) or croak 'usage: $font->writeVF([VFFILE, TFMFILE]);';
    my ($metrics, $vffile, $tfmfile) = @_;
    my $fontid = $metrics->fontId;
    
    croak "'$fontid' is not a virtual font" unless $metrics->{MapInfo};
    my $vplfile = "$fontid.vpl";
    $vffile     = "$fontid.vf"  unless defined $vffile;
    $tfmfile    = "$fontid.tfm" unless defined $tfmfile;
    writePropertyList($metrics, $vplfile);
    my $cmd = "vptovf $vplfile $vffile $tfmfile";
    my $diagnostics =  `$cmd`;
    croak "$cmd failed" unless defined $diagnostics;
    $diagnostics =~ s/^/\tvptovf: /mg if $diagnostics ne '';
    $diagnostics =~ s/I had to round/rounded/mg;
    print STDERR $diagnostics;
    unlink $vplfile unless -e '.keep_plfiles';
    return $metrics;
}

1;    
